#   Copyright 2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""Keypair action implementations"""

import argparse
import collections
from collections.abc import Iterable, Sequence
import logging
import os
from typing import Any

from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives import serialization
from openstack.compute.v2 import keypair as _keypair
from openstack import resource
from openstack import utils as sdk_utils
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common


LOG = logging.getLogger(__name__)
Keypair = collections.namedtuple('Keypair', 'private_key public_key')


def _generate_keypair() -> Keypair:
    """Generate a Ed25519 keypair in OpenSSH format.

    :returns: A `Keypair` named tuple with the generated private and public
    keys.
    """
    key = ed25519.Ed25519PrivateKey.generate()
    private_key = key.private_bytes(
        serialization.Encoding.PEM,
        serialization.PrivateFormat.OpenSSH,
        serialization.NoEncryption(),
    ).decode()
    public_key = (
        key.public_key()
        .public_bytes(
            serialization.Encoding.OpenSSH, serialization.PublicFormat.OpenSSH
        )
        .decode()
    )

    return Keypair(private_key, public_key)


def _get_keypair_columns(
    item: resource.Resource,
    hide_pub_key: bool = False,
    hide_priv_key: bool = False,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    # To maintain backwards compatibility we need to rename sdk props to
    # whatever OSC was using before
    hidden_columns = ['links', 'location']
    if hide_pub_key:
        hidden_columns.append('public_key')
    if hide_priv_key:
        hidden_columns.append('private_key')
    return utils.get_osc_show_columns_for_sdk_resource(
        item, {}, hidden_columns
    )


class CreateKeypair(command.ShowOne):
    _description = _("Create new public or private key for server ssh access")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name', metavar='<name>', help=_("New public or private key name")
        )
        key_group = parser.add_mutually_exclusive_group()
        key_group.add_argument(
            '--public-key',
            metavar='<file>',
            help=_(
                "Filename for public key to add. "
                "If not used, generates a private key in ssh-ed25519 format. "
                "To generate keys in other formats, including the legacy "
                "ssh-rsa format, you must use an external tool such as "
                "ssh-keygen and specify this argument."
            ),
        )
        key_group.add_argument(
            '--private-key',
            metavar='<file>',
            help=_(
                "Filename for private key to save. "
                "If not used, print private key in console."
            ),
        )
        parser.add_argument(
            '--type',
            metavar='<type>',
            choices=['ssh', 'x509'],
            help=_(
                'Keypair type '
                '(supported by --os-compute-api-version 2.2 or above)'
            ),
        )
        parser.add_argument(
            '--user',
            metavar='<user>',
            help=_(
                'The owner of the keypair (admin only) (name or ID) '
                '(supported by --os-compute-api-version 2.10 or above)'
            ),
        )
        identity_common.add_user_domain_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        compute_client = self.app.client_manager.compute
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        kwargs = {'name': parsed_args.name}

        if parsed_args.public_key:
            try:
                with open(os.path.expanduser(parsed_args.public_key)) as p:
                    public_key = p.read()
            except OSError as e:
                msg = _("Key file %(public_key)s not found: %(exception)s")
                raise exceptions.CommandError(
                    msg
                    % {
                        "public_key": parsed_args.public_key,
                        "exception": e,
                    }
                )

            kwargs['public_key'] = public_key
        else:
            generated_keypair = _generate_keypair()
            kwargs['public_key'] = generated_keypair.public_key

            # If user have us a file, save private key into specified file
            if parsed_args.private_key:
                try:
                    with open(
                        os.path.expanduser(parsed_args.private_key), 'w+'
                    ) as p:
                        p.write(generated_keypair.private_key)
                except OSError as e:
                    msg = _(
                        "Key file %(private_key)s can not be saved: "
                        "%(exception)s"
                    )
                    raise exceptions.CommandError(
                        msg
                        % {
                            "private_key": parsed_args.private_key,
                            "exception": e,
                        }
                    )

        if parsed_args.type:
            if not sdk_utils.supports_microversion(compute_client, '2.2'):
                msg = _(
                    '--os-compute-api-version 2.2 or greater is required to '
                    'support the --type option'
                )
                raise exceptions.CommandError(msg)

            kwargs['key_type'] = parsed_args.type

        if parsed_args.user:
            if not sdk_utils.supports_microversion(compute_client, '2.10'):
                msg = _(
                    '--os-compute-api-version 2.10 or greater is required to '
                    'support the --user option'
                )
                raise exceptions.CommandError(msg)

            kwargs['user_id'] = identity_common.find_user_id_sdk(
                identity_client,
                parsed_args.user,
                parsed_args.user_domain,
            )

        keypair = compute_client.create_keypair(**kwargs)

        # NOTE(dtroyer): how do we want to handle the display of the private
        #                key when it needs to be communicated back to the user
        #                For now, duplicate nova keypair-add command output
        if parsed_args.public_key or parsed_args.private_key:
            display_columns, columns = _get_keypair_columns(
                keypair, hide_pub_key=True, hide_priv_key=True
            )
            data = utils.get_item_properties(keypair, columns)

            return (display_columns, data)
        else:
            self.app.stdout.write(generated_keypair.private_key)
            return ((), ())


class DeleteKeypair(command.Command):
    _description = _("Delete public or private key(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<key>',
            nargs='+',
            help=_("Name of key(s) to delete (name only)"),
        )
        parser.add_argument(
            '--user',
            metavar='<user>',
            help=_(
                'The owner of the keypair. (admin only) (name or ID). '
                'Requires ``--os-compute-api-version`` 2.10 or greater.'
            ),
        )
        identity_common.add_user_domain_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> Any:
        compute_client = self.app.client_manager.compute
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        kwargs = {}
        result = 0

        if parsed_args.user:
            if not sdk_utils.supports_microversion(compute_client, '2.10'):
                msg = _(
                    '--os-compute-api-version 2.10 or greater is required to '
                    'support the --user option'
                )
                raise exceptions.CommandError(msg)

            kwargs['user_id'] = identity_common.find_user_id_sdk(
                identity_client,
                parsed_args.user,
                parsed_args.user_domain,
            )

        for n in parsed_args.name:
            try:
                compute_client.delete_keypair(
                    n, **kwargs, ignore_missing=False
                )
            except Exception as e:
                result += 1
                LOG.error(
                    _("Failed to delete key with name '%(name)s': %(e)s"),
                    {'name': n, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.name)
            msg = _("%(result)s of %(total)s keys failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListKeypair(command.Lister):
    _description = _("List key fingerprints")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        user_group = parser.add_mutually_exclusive_group()
        user_group.add_argument(
            '--user',
            metavar='<user>',
            help=_(
                'Show keypairs for another user (admin only) (name or ID). '
                'Requires ``--os-compute-api-version`` 2.10 or greater.'
            ),
        )
        identity_common.add_user_domain_option_to_parser(parser)
        user_group.add_argument(
            '--project',
            metavar='<project>',
            help=_(
                'Show keypairs for all users associated with project '
                '(admin only) (name or ID). '
                'Requires ``--os-compute-api-version`` 2.10 or greater.'
            ),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        pagination.add_marker_pagination_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        compute_client = self.app.client_manager.compute
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        kwargs = {}

        if parsed_args.marker:
            if not sdk_utils.supports_microversion(compute_client, '2.35'):
                msg = _(
                    '--os-compute-api-version 2.35 or greater is required '
                    'to support the --marker option'
                )
                raise exceptions.CommandError(msg)

            kwargs['marker'] = parsed_args.marker

        if parsed_args.limit:
            if not sdk_utils.supports_microversion(compute_client, '2.35'):
                msg = _(
                    '--os-compute-api-version 2.35 or greater is required '
                    'to support the --limit option'
                )
                raise exceptions.CommandError(msg)

            kwargs['limit'] = parsed_args.limit
        if parsed_args.max_items is not None:
            kwargs['max_items'] = parsed_args.max_items

        data: list[_keypair.Keypair]
        if parsed_args.project:
            if not sdk_utils.supports_microversion(compute_client, '2.10'):
                msg = _(
                    '--os-compute-api-version 2.10 or greater is required to '
                    'support the --project option'
                )
                raise exceptions.CommandError(msg)

            if parsed_args.marker:
                # NOTE(stephenfin): Because we're doing this client-side, we
                # can't really rely on the marker, because we don't know what
                # user the marker is associated with
                msg = _('--project is not compatible with --marker')

            # NOTE(stephenfin): This is done client side because nova doesn't
            # currently support doing so server-side. If this is slow, we can
            # think about spinning up a threadpool or similar.
            project = identity_common.find_project_id_sdk(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            )
            assignments = identity_client.role_assignments(
                scope_project_id=project
            )
            user_ids = set()
            for assignment in assignments:
                if assignment.user:
                    user_ids.add(assignment.user['id'])

            data = []
            for user_id in user_ids:
                kwargs['user_id'] = user_id
                data.extend(compute_client.keypairs(**kwargs))
        elif parsed_args.user:
            if not sdk_utils.supports_microversion(compute_client, '2.10'):
                msg = _(
                    '--os-compute-api-version 2.10 or greater is required to '
                    'support the --user option'
                )
                raise exceptions.CommandError(msg)

            user_id = identity_common.find_user_id_sdk(
                identity_client,
                parsed_args.user,
                parsed_args.user_domain,
            )
            kwargs['user_id'] = user_id

            data = list(compute_client.keypairs(**kwargs))
        else:
            data = list(compute_client.keypairs(**kwargs))

        columns: tuple[str, ...] = ("Name", "Fingerprint")

        if sdk_utils.supports_microversion(compute_client, '2.2'):
            columns += ("Type",)

        return (
            columns,
            (utils.get_item_properties(s, columns) for s in data),
        )


class ShowKeypair(command.ShowOne):
    _description = _("Display key details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<key>',
            help=_("Public or private key to display (name only)"),
        )
        parser.add_argument(
            '--public-key',
            action='store_true',
            default=False,
            help=_("Show only bare public key paired with the generated key"),
        )
        parser.add_argument(
            '--user',
            metavar='<user>',
            help=_(
                'The owner of the keypair. (admin only) (name or ID). '
                'Requires ``--os-compute-api-version`` 2.10 or greater.'
            ),
        )
        identity_common.add_user_domain_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        compute_client = self.app.client_manager.compute
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        kwargs = {}

        if parsed_args.user:
            if not sdk_utils.supports_microversion(compute_client, '2.10'):
                msg = _(
                    '--os-compute-api-version 2.10 or greater is required to '
                    'support the --user option'
                )
                raise exceptions.CommandError(msg)

            kwargs['user_id'] = identity_common.find_user_id_sdk(
                identity_client,
                parsed_args.user,
                parsed_args.user_domain,
            )

        keypair = compute_client.find_keypair(
            parsed_args.name, **kwargs, ignore_missing=False
        )

        if not parsed_args.public_key:
            display_columns, columns = _get_keypair_columns(
                keypair, hide_pub_key=True
            )
            data = utils.get_item_properties(keypair, columns)
            return (display_columns, data)
        else:
            self.app.stdout.write(keypair.public_key)
            return ((), ())



# Copyright (c) 2016 Juniper Networks Inc.
# All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from openstack.network.v2 import bgpvpn as _bgpvpn
from osc_lib.cli import format_columns
from osc_lib.cli import identity as osc_id
from osc_lib.cli.parseractions import KeyValueAction
from osc_lib import exceptions
from osc_lib import utils as osc_utils
from osc_lib.utils import columns as column_util

from openstackclient import command
from openstackclient.i18n import _

LOG = logging.getLogger(__name__)

_attr_map = (
    ('id', 'ID', column_util.LIST_BOTH),
    ('project_id', 'Project', column_util.LIST_LONG_ONLY),
    ('name', 'Name', column_util.LIST_BOTH),
    ('type', 'Type', column_util.LIST_BOTH),
    ('route_targets', 'Route Targets', column_util.LIST_LONG_ONLY),
    ('import_targets', 'Import Targets', column_util.LIST_LONG_ONLY),
    ('export_targets', 'Export Targets', column_util.LIST_LONG_ONLY),
    (
        'route_distinguishers',
        'Route Distinguishers',
        column_util.LIST_LONG_ONLY,
    ),
    ('networks', 'Associated Networks', column_util.LIST_LONG_ONLY),
    ('routers', 'Associated Routers', column_util.LIST_LONG_ONLY),
    ('ports', 'Associated Ports', column_util.LIST_LONG_ONLY),
    ('vni', 'VNI', column_util.LIST_LONG_ONLY),
    ('local_pref', 'Local Pref', column_util.LIST_LONG_ONLY),
)
_formatters = {
    'route_targets': format_columns.ListColumn,
    'import_targets': format_columns.ListColumn,
    'export_targets': format_columns.ListColumn,
    'route_distinguishers': format_columns.ListColumn,
    'networks': format_columns.ListColumn,
    'routers': format_columns.ListColumn,
    'ports': format_columns.ListColumn,
}


def _get_columns(
    item: _bgpvpn.BgpVpn,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    column_map: dict[str, str] = {}
    hidden_columns = ['location', 'tenant_id']
    return osc_utils.get_osc_show_columns_for_sdk_resource(
        item, column_map, hidden_columns
    )


def _get_common_parser(
    parser: argparse.ArgumentParser, update: str | None = None
) -> None:
    """Adds to parser arguments common to create, set and unset commands.

    :params ArgumentParser parser: argparse object contains all command's
        arguments
    :params string update: Determines if it is a create command (value: None),
        it is a set command (value: 'set') or if it is an unset command (value:
        'unset')
    """
    ADD_RT = _("Add Route Target to import/export list")
    REMOVE_RT = _("Remove Route Target from import/export list")
    ADD_IMPORT_RT = _("Add Route Target to import list")
    DEL_IMPORT_RT = _("Remove Route Target from import list")
    ADD_EXPORT_RT = _("Add Route Target to export list")
    DEL_EXPORT_RT = _("Remove Route Target from export list")
    ADD_RD = _(
        "Add Route Distinguisher to the list of Route Distinguishers "
        "from which a Route Distinguishers will be picked from to "
        "advertise a VPN route"
    )
    REMOVE_RD = _(
        "Remove Route Distinguisher from the list of Route "
        "Distinguishers from which a Route Distinguishers will be "
        "picked from to advertise a VPN route"
    )
    REPEAT_RT = _("repeat option for multiple Route Targets")
    REPEAT_RD = _("repeat option for multiple Route Distinguishers")

    def is_appended() -> bool:
        return update is None or update == 'set'

    if update is None or update == 'set':
        parser.add_argument(
            '--name',
            metavar="<name>",
            help=_("Name of the BGP VPN"),
        )
    parser.add_argument(
        '--route-target',
        dest='route_targets',
        action='append',
        metavar="<route-target>",
        help=f"{ADD_RT if is_appended() else REMOVE_RT} ({REPEAT_RT})",
    )
    if update:
        parser.add_argument(
            '--no-route-target' if update == 'set' else '--all-route-target',
            dest='purge_route_target',
            action='store_true',
            help=_('Empty route target list'),
        )
    import_target_action = ADD_IMPORT_RT if is_appended() else DEL_IMPORT_RT
    parser.add_argument(
        '--import-target',
        dest='import_targets',
        action='append',
        metavar="<import-target>",
        help=f"{import_target_action} ({REPEAT_RT})",
    )
    if update:
        parser.add_argument(
            '--no-import-target' if update == 'set' else '--all-import-target',
            dest='purge_import_target',
            action='store_true',
            help=_('Empty import route target list'),
        )
    export_target_action = ADD_EXPORT_RT if is_appended() else DEL_EXPORT_RT
    parser.add_argument(
        '--export-target',
        dest='export_targets',
        action='append',
        metavar="<export-target>",
        help=f"{export_target_action} ({REPEAT_RT})",
    )
    if update:
        parser.add_argument(
            '--no-export-target' if update == 'set' else '--all-export-target',
            dest='purge_export_target',
            action='store_true',
            help=_('Empty export route target list'),
        )
    parser.add_argument(
        '--route-distinguisher',
        dest='route_distinguishers',
        action='append',
        metavar="<route-distinguisher>",
        help=f"{ADD_RD if is_appended() else REMOVE_RD} ({REPEAT_RD})",
    )
    if update:
        parser.add_argument(
            '--no-route-distinguisher'
            if update == 'set'
            else '--all-route-distinguisher',
            dest='purge_route_distinguisher',
            action='store_true',
            help=_('Empty route distinguisher list'),
        )
    parser.add_argument(
        '--vni',
        type=int,
        help=_(
            'VXLAN Network Identifier to be used for this BGPVPN '
            'when a VXLAN encapsulation is used'
        ),
    )
    parser.add_argument(
        '--local-pref',
        type=int,
        dest='local_pref',
        help=_(
            'Default BGP LOCAL_PREF to use in route advertisements'
            'towards this BGPVPN.'
        ),
    )


def _args2body(
    client_manager: Any, id: str, action: str, args: argparse.Namespace
) -> dict[str, Any]:

    if not (
        args.purge_route_target
        and args.purge_import_target
        and args.purge_export_target
        and args.purge_route_distinguisher
    ) and (
        args.route_targets
        or args.import_targets
        or args.export_targets
        or args.route_distinguishers
    ):
        bgpvpn = client_manager.network.get_bgpvpn(id)

    attrs: dict[str, Any] = {}

    if 'name' in args and args.name is not None:
        attrs['name'] = str(args.name)

    if 'vni' in args and args.vni is not None:
        attrs['vni'] = args.vni

    if 'local_pref' in args and args.local_pref is not None:
        attrs['local_pref'] = args.local_pref

    if args.purge_route_target:
        attrs['route_targets'] = []
    elif args.route_targets:
        if action == 'set':
            attrs['route_targets'] = list(
                set(bgpvpn['route_targets']) | set(args.route_targets)
            )
        elif action == 'unset':
            attrs['route_targets'] = list(
                set(bgpvpn['route_targets']) - set(args.route_targets)
            )

    if args.purge_import_target:
        attrs['import_targets'] = []
    elif args.import_targets:
        if action == 'set':
            attrs['import_targets'] = list(
                set(bgpvpn['import_targets']) | set(args.import_targets)
            )
        elif action == 'unset':
            attrs['import_targets'] = list(
                set(bgpvpn['import_targets']) - set(args.import_targets)
            )

    if args.purge_export_target:
        attrs['export_targets'] = []
    elif args.export_targets:
        if action == 'set':
            attrs['export_targets'] = list(
                set(bgpvpn['export_targets']) | set(args.export_targets)
            )
        elif action == 'unset':
            attrs['export_targets'] = list(
                set(bgpvpn['export_targets']) - set(args.export_targets)
            )

    if args.purge_route_distinguisher:
        attrs['route_distinguishers'] = []
    elif args.route_distinguishers:
        if action == 'set':
            attrs['route_distinguishers'] = list(
                set(bgpvpn['route_distinguishers'])
                | set(args.route_distinguishers)
            )
        elif action == 'unset':
            attrs['route_distinguishers'] = list(
                set(bgpvpn['route_distinguishers'])
                - set(args.route_distinguishers)
            )

    return attrs


class CreateBgpvpn(command.ShowOne):
    _description = _("Create BGP VPN resource")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        osc_id.add_project_owner_option_to_parser(parser)
        _get_common_parser(parser)
        parser.add_argument(
            '--type',
            default='l3',
            choices=['l2', 'l3'],
            help=_(
                "BGP VPN type selection between IP VPN (l3) and Ethernet "
                "VPN (l2) (default: l3)"
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        attrs = {}
        if parsed_args.name is not None:
            attrs['name'] = str(parsed_args.name)
        if parsed_args.type is not None:
            attrs['type'] = parsed_args.type
        if parsed_args.route_targets is not None:
            attrs['route_targets'] = parsed_args.route_targets
        if parsed_args.import_targets is not None:
            attrs['import_targets'] = parsed_args.import_targets
        if parsed_args.export_targets is not None:
            attrs['export_targets'] = parsed_args.export_targets
        if parsed_args.route_distinguishers is not None:
            attrs['route_distinguishers'] = parsed_args.route_distinguishers
        if parsed_args.vni is not None:
            attrs['vni'] = parsed_args.vni
        if parsed_args.local_pref is not None:
            attrs['local_pref'] = parsed_args.local_pref
        if 'project' in parsed_args and parsed_args.project is not None:
            project_id = osc_id.find_project(
                self.app.client_manager.sdk_connection,
                parsed_args.project,
                parsed_args.project_domain,
            ).id
            attrs['project_id'] = project_id
        obj = client.create_bgpvpn(**attrs)
        display_columns, columns = _get_columns(obj)
        data = osc_utils.get_dict_properties(
            obj, columns, formatters=_formatters
        )
        return display_columns, data


class SetBgpvpn(command.Command):
    _description = _("Set BGP VPN properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'bgpvpn',
            metavar="<bgpvpn>",
            help=_("BGP VPN to update (name or ID)"),
        )
        _get_common_parser(parser, update='set')
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        id = client.find_bgpvpn(parsed_args.bgpvpn, ignore_missing=False)['id']
        body = _args2body(self.app.client_manager, id, 'set', parsed_args)
        client.update_bgpvpn(id, **body)


class UnsetBgpvpn(command.Command):
    _description = _("Unset BGP VPN properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'bgpvpn',
            metavar="<bgpvpn>",
            help=_("BGP VPN to update (name or ID)"),
        )
        _get_common_parser(parser, update='unset')
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        id = client.find_bgpvpn(parsed_args.bgpvpn, ignore_missing=False)['id']
        body = _args2body(self.app.client_manager, id, 'unset', parsed_args)
        client.update_bgpvpn(id, **body)


class DeleteBgpvpn(command.Command):
    _description = _("Delete BGP VPN resource(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'bgpvpns',
            metavar="<bgpvpn>",
            nargs="+",
            help=_("BGP VPN(s) to delete (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        fails = 0
        for id_or_name in parsed_args.bgpvpns:
            try:
                id = client.find_bgpvpn(id_or_name, ignore_missing=False)['id']
                client.delete_bgpvpn(id)
                LOG.warning("BGP VPN %(id)s deleted", {'id': id})
            except Exception as e:
                fails += 1
                LOG.error(
                    "Failed to delete BGP VPN with name or ID "
                    "'%(id_or_name)s': %(e)s",
                    {'id_or_name': id_or_name, 'e': e},
                )
        if fails > 0:
            msg = _("Failed to delete %(fails)s of %(total)s BGP VPN.") % {
                'fails': fails,
                'total': len(parsed_args.bgpvpns),
            }
            raise exceptions.CommandError(msg)


class ListBgpvpn(command.Lister):
    _description = _("List BGP VPN resources")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        osc_id.add_project_owner_option_to_parser(parser)
        parser.add_argument(
            '--long',
            action='store_true',
            help=_("List additional fields in output"),
        )
        parser.add_argument(
            '--property',
            metavar="<key=value>",
            default=dict(),
            help=_(
                "Filter property to apply on returned BGP VPNs (repeat to "
                "filter on multiple properties)"
            ),
            action=KeyValueAction,
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        client = self.app.client_manager.network
        params = {}
        if parsed_args.project is not None:
            project_id = osc_id.find_project(
                self.app.client_manager.sdk_connection,
                parsed_args.project,
                parsed_args.project_domain,
            ).id
            params['project_id'] = project_id
        if parsed_args.property:
            params.update(parsed_args.property)
        objs = client.bgpvpns(**params)
        headers, columns = column_util.get_column_definitions(
            list(_attr_map), long_listing=parsed_args.long
        )
        return (
            headers,
            (
                osc_utils.get_dict_properties(
                    s, columns, formatters=_formatters
                )
                for s in objs
            ),
        )


class ShowBgpvpn(command.ShowOne):
    _description = _("Show information of a given BGP VPN")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'bgpvpn',
            metavar="<bgpvpn>",
            help=_("BGP VPN to display (name or ID)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        id = client.find_bgpvpn(parsed_args.bgpvpn, ignore_missing=False)['id']
        obj = client.get_bgpvpn(id)
        display_columns, columns = _get_columns(obj)
        data = osc_utils.get_dict_properties(
            obj, columns, formatters=_formatters
        )
        return display_columns, data



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from openstackclient.tests.functional.network.v2 import common


class L3NDPProxyTests(common.NetworkTests):
    def setUp(self):
        super().setUp()

        if not self.is_extension_enabled('l3-ndp-proxy'):
            self.skipTest("No l3-ndp-proxy extension present")

        self.ROT_NAME = self.getUniqueString()
        self.EXT_NET_NAME = self.getUniqueString()
        self.EXT_SUB_NAME = self.getUniqueString()
        self.INT_NET_NAME = self.getUniqueString()
        self.INT_SUB_NAME = self.getUniqueString()
        self.INT_PORT_NAME = self.getUniqueString()
        self.ADDR_SCOPE_NAME = self.getUniqueString()
        self.SUBNET_P_NAME = self.getUniqueString()
        self.created_ndp_proxies = []

        json_output = self.openstack(
            f'address scope create --ip-version 6 {self.ADDR_SCOPE_NAME}',
            parse_output=True,
        )
        self.assertIsNotNone(json_output['id'])
        self.ADDRESS_SCOPE_ID = json_output['id']
        json_output = self.openstack(
            f'subnet pool create {self.SUBNET_P_NAME} '
            f'--address-scope {self.ADDRESS_SCOPE_ID} '
            '--pool-prefix 2001:db8::/96 --default-prefix-length 112',
            parse_output=True,
        )
        self.assertIsNotNone(json_output['id'])
        self.SUBNET_POOL_ID = json_output['id']
        json_output = self.openstack(
            'network create --external ' + self.EXT_NET_NAME,
            parse_output=True,
        )
        self.assertIsNotNone(json_output['id'])
        self.EXT_NET_ID = json_output['id']
        json_output = self.openstack(
            'subnet create --ip-version 6 --subnet-pool '
            f'{self.SUBNET_POOL_ID} --network {self.EXT_NET_ID} {self.EXT_SUB_NAME}',
            parse_output=True,
        )
        self.assertIsNotNone(json_output['id'])
        self.EXT_SUB_ID = json_output['id']
        json_output = self.openstack(
            'router create ' + self.ROT_NAME,
            parse_output=True,
        )
        self.assertIsNotNone(json_output['id'])
        self.ROT_ID = json_output['id']
        output = self.openstack(
            f'router set {self.ROT_ID} --external-gateway {self.EXT_NET_ID}'
        )
        self.assertEqual('', output)
        output = self.openstack('router set --enable-ndp-proxy ' + self.ROT_ID)
        self.assertEqual('', output)
        json_output = self.openstack(
            'router show -c enable_ndp_proxy ' + self.ROT_ID,
            parse_output=True,
        )
        self.assertTrue(json_output['enable_ndp_proxy'])
        json_output = self.openstack(
            'network create ' + self.INT_NET_NAME,
            parse_output=True,
        )
        self.assertIsNotNone(json_output['id'])
        self.INT_NET_ID = json_output['id']
        json_output = self.openstack(
            'subnet create --ip-version 6 --subnet-pool '
            f'{self.SUBNET_POOL_ID} --network {self.INT_NET_ID} {self.INT_SUB_NAME}',
            parse_output=True,
        )
        self.assertIsNotNone(json_output['id'])
        self.INT_SUB_ID = json_output['id']
        json_output = self.openstack(
            f'port create --network {self.INT_NET_ID} {self.INT_PORT_NAME}',
            parse_output=True,
        )
        self.assertIsNotNone(json_output['id'])
        self.INT_PORT_ID = json_output['id']
        self.INT_PORT_ADDRESS = json_output['fixed_ips'][0]['ip_address']
        output = self.openstack(
            'router add subnet ' + self.ROT_ID + ' ' + self.INT_SUB_ID
        )
        self.assertEqual('', output)

    def tearDown(self):
        for ndp_proxy in self.created_ndp_proxies:
            output = self.openstack(
                'router ndp proxy delete ' + ndp_proxy['id']
            )
            self.assertEqual('', output)
        output = self.openstack('port delete ' + self.INT_PORT_ID)
        self.assertEqual('', output)
        output = self.openstack(
            'router set --disable-ndp-proxy ' + self.ROT_ID
        )
        self.assertEqual('', output)
        output = self.openstack(
            'router remove subnet ' + self.ROT_ID + ' ' + self.INT_SUB_ID
        )
        self.assertEqual('', output)
        output = self.openstack('subnet delete ' + self.INT_SUB_ID)
        self.assertEqual('', output)
        output = self.openstack('network delete ' + self.INT_NET_ID)
        self.assertEqual('', output)
        output = self.openstack(
            'router unset ' + self.ROT_ID + ' ' + '--external-gateway'
        )
        self.assertEqual('', output)
        output = self.openstack('router delete ' + self.ROT_ID)
        self.assertEqual('', output)
        output = self.openstack('subnet delete ' + self.EXT_SUB_ID)
        self.assertEqual('', output)
        output = self.openstack('network delete ' + self.EXT_NET_ID)
        self.assertEqual('', output)
        output = self.openstack('subnet pool delete ' + self.SUBNET_POOL_ID)
        self.assertEqual('', output)
        output = self.openstack(
            'address scope delete ' + self.ADDRESS_SCOPE_ID
        )
        self.assertEqual('', output)
        super().tearDown()

    def _create_ndp_proxies(self, ndp_proxies):
        for ndp_proxy in ndp_proxies:
            output = self.openstack(
                'router ndp proxy create {router} --name {name} '
                '--port {port} --ip-address {address}'.format(
                    router=ndp_proxy['router_id'],
                    name=ndp_proxy['name'],
                    port=ndp_proxy['port_id'],
                    address=ndp_proxy['address'],
                ),
                parse_output=True,
            )
            self.assertEqual(ndp_proxy['router_id'], output['router_id'])
            self.assertEqual(ndp_proxy['port_id'], output['port_id'])
            self.assertEqual(ndp_proxy['address'], output['ip_address'])
            self.created_ndp_proxies.append(output)

    def test_create_ndp_proxy(self):
        ndp_proxies = [
            {
                'name': self.getUniqueString(),
                'router_id': self.ROT_ID,
                'port_id': self.INT_PORT_ID,
                'address': self.INT_PORT_ADDRESS,
            }
        ]
        self._create_ndp_proxies(ndp_proxies)

    def test_ndp_proxy_list(self):
        ndp_proxies = {
            'name': self.getUniqueString(),
            'router_id': self.ROT_ID,
            'port_id': self.INT_PORT_ID,
            'address': self.INT_PORT_ADDRESS,
        }
        self._create_ndp_proxies([ndp_proxies])
        ndp_proxy = self.openstack(
            'router ndp proxy list',
            parse_output=True,
        )[0]
        self.assertEqual(ndp_proxies['name'], ndp_proxy['Name'])
        self.assertEqual(ndp_proxies['router_id'], ndp_proxy['Router ID'])
        self.assertEqual(ndp_proxies['address'], ndp_proxy['IP Address'])

    def test_ndp_proxy_set_and_show(self):
        ndp_proxies = {
            'name': self.getUniqueString(),
            'router_id': self.ROT_ID,
            'port_id': self.INT_PORT_ID,
            'address': self.INT_PORT_ADDRESS,
        }
        description = 'balala'
        self._create_ndp_proxies([ndp_proxies])
        ndp_proxy_id = self.created_ndp_proxies[0]['id']
        output = self.openstack(
            f'router ndp proxy set --description {description} {ndp_proxy_id}'
        )
        self.assertEqual('', output)
        json_output = self.openstack(
            'router ndp proxy show ' + ndp_proxy_id,
            parse_output=True,
        )
        self.assertEqual(ndp_proxies['name'], json_output['name'])
        self.assertEqual(ndp_proxies['router_id'], json_output['router_id'])
        self.assertEqual(ndp_proxies['port_id'], json_output['port_id'])
        self.assertEqual(ndp_proxies['address'], json_output['ip_address'])
        self.assertEqual(description, json_output['description'])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
from collections.abc import Iterable, Sequence
from typing import Any

from osc_lib.cli import identity as identity_utils
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common
from openstackclient.network import utils as network_utils


MIN_AS_NUM = 1
MAX_AS_NUM = 4294967295


def _get_attrs(parsed_args: argparse.Namespace) -> dict[str, Any]:
    attrs = {}

    # Validate password
    if 'auth_type' in parsed_args:
        if parsed_args.auth_type != 'none':
            if 'password' not in parsed_args or parsed_args.password is None:
                raise exceptions.CommandError(
                    _('Must provide password if auth-type is specified.')
                )
        if (
            parsed_args.auth_type == 'none'
            and parsed_args.password is not None
        ):
            raise exceptions.CommandError(
                _('Must provide auth-type if password is specified.')
            )
        attrs['auth_type'] = parsed_args.auth_type

    if parsed_args.name is not None:
        attrs['name'] = parsed_args.name
    if 'remote_as' in parsed_args:
        attrs['remote_as'] = parsed_args.remote_as
    if 'peer_ip' in parsed_args:
        attrs['peer_ip'] = parsed_args.peer_ip
    if 'password' in parsed_args:
        attrs['password'] = parsed_args.password
    return attrs


class CreateBgpPeer(command.ShowOne):
    _description = _("Create a BGP peer")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name', metavar='<name>', help=_("Name of the BGP peer to create")
        )
        parser.add_argument(
            '--peer-ip',
            metavar='<peer-ip-address>',
            required=True,
            help=_("Peer IP address"),
        )
        parser.add_argument(
            '--remote-as',
            required=True,
            metavar='<peer-remote-as>',
            help=_(
                "Peer AS number. (Integer in [%(min_val)s, %(max_val)s] "
                "is allowed)"
            )
            % {
                'min_val': MIN_AS_NUM,
                'max_val': MAX_AS_NUM,
            },
        )
        parser.add_argument(
            '--auth-type',
            metavar='<peer-auth-type>',
            choices=['none', 'md5'],
            type=network_utils.convert_to_lowercase,
            default='none',
            help=_(
                "Authentication algorithm. Supported algorithms: "
                "none (default), md5"
            ),
        )
        parser.add_argument(
            '--password',
            metavar='<auth-password>',
            help=_("Authentication password"),
        )
        identity_utils.add_project_owner_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network

        attrs = _get_attrs(parsed_args)
        if parsed_args.project is not None:
            identity_client = self.app.client_manager.sdk_connection.identity
            project_id = identity_common.find_project_id_sdk(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            )
            attrs['project_id'] = project_id

        obj = client.create_bgp_peer(**attrs)
        display_columns, columns = utils.get_osc_show_columns_for_sdk_resource(
            obj, {}, ['location', 'tenant_id']
        )
        data = utils.get_dict_properties(obj, columns)
        return display_columns, data


class DeleteBgpPeer(command.Command):
    _description = _("Delete a BGP peer")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'bgp_peer',
            metavar="<bgp-peer>",
            help=_("BGP peer to delete (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        id = client.find_bgp_peer(
            parsed_args.bgp_peer, ignore_missing=False
        ).id
        client.delete_bgp_peer(id)


class ListBgpPeer(command.Lister):
    _description = _("List BGP peers")

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        data = self.app.client_manager.network.bgp_peers(retrieve_all=True)
        headers = ('ID', 'Name', 'Peer IP', 'Remote AS')
        columns = ('id', 'name', 'peer_ip', 'remote_as')
        return (
            headers,
            (
                utils.get_dict_properties(
                    s,
                    columns,
                )
                for s in data
            ),
        )


class SetBgpPeer(command.Command):
    _description = _("Update a BGP peer")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument('--name', help=_("Updated name of the BGP peer"))
        parser.add_argument(
            '--password',
            metavar='<auth-password>',
            help=_("Updated authentication password"),
        )
        parser.add_argument(
            'bgp_peer',
            metavar="<bgp-peer>",
            help=_("BGP peer to update (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        id = client.find_bgp_peer(
            parsed_args.bgp_peer, ignore_missing=False
        ).id
        attrs = _get_attrs(parsed_args)
        client.update_bgp_peer(id, **attrs)


class ShowBgpPeer(command.ShowOne):
    _description = _("Show information for a BGP peer")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'bgp_peer',
            metavar="<bgp-peer>",
            help=_("BGP peer to display (name or ID)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        obj = client.find_bgp_peer(parsed_args.bgp_peer, ignore_missing=False)
        display_columns, columns = utils.get_osc_show_columns_for_sdk_resource(
            obj, {}, ['location', 'tenant_id']
        )
        data = utils.get_dict_properties(obj, columns)
        return display_columns, data



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock

from keystoneauth1 import exceptions as ks_exc
from osc_lib import exceptions
from osc_lib import utils

from openstackclient.identity.v2_0 import user
from openstackclient.tests.unit.identity.v2_0 import fakes as identity_fakes


class TestUser(identity_fakes.TestIdentity):
    fake_project = identity_fakes.FakeProject.create_one_project()
    attr = {
        'tenantId': fake_project.id,
    }
    fake_user = identity_fakes.FakeUser.create_one_user(attr)

    def setUp(self):
        super().setUp()

        # Get a shortcut to the TenantManager Mock
        self.projects_mock = self.identity_client.tenants
        self.projects_mock.reset_mock()

        # Get a shortcut to the UserManager Mock
        self.users_mock = self.identity_client.users
        self.users_mock.reset_mock()


class TestUserCreate(TestUser):
    fake_project_c = identity_fakes.FakeProject.create_one_project()
    attr = {
        'tenantId': fake_project_c.id,
    }
    fake_user_c = identity_fakes.FakeUser.create_one_user(attr)

    columns = (
        'email',
        'enabled',
        'id',
        'name',
        'project_id',
    )
    datalist = (
        fake_user_c.email,
        True,
        fake_user_c.id,
        fake_user_c.name,
        fake_project_c.id,
    )

    def setUp(self):
        super().setUp()

        self.projects_mock.get.return_value = self.fake_project_c

        self.users_mock.create.return_value = self.fake_user_c

        # Get the command object to test
        self.cmd = user.CreateUser(self.app, None)

    def test_user_create_no_options(self):
        arglist = [
            self.fake_user_c.name,
        ]
        verifylist = [
            ('enable', False),
            ('disable', False),
            ('name', self.fake_user_c.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'enabled': True,
            'tenant_id': None,
        }
        # UserManager.create(name, password, email, tenant_id=, enabled=)
        self.users_mock.create.assert_called_with(
            self.fake_user_c.name, None, None, **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_password(self):
        arglist = [
            '--password',
            'secret',
            self.fake_user_c.name,
        ]
        verifylist = [
            ('name', self.fake_user_c.name),
            ('password_prompt', False),
            ('password', 'secret'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'enabled': True,
            'tenant_id': None,
        }
        # UserManager.create(name, password, email, tenant_id=, enabled=)
        self.users_mock.create.assert_called_with(
            self.fake_user_c.name, 'secret', None, **kwargs
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_password_prompt(self):
        arglist = [
            '--password-prompt',
            self.fake_user_c.name,
        ]
        verifylist = [
            ('name', self.fake_user_c.name),
            ('password_prompt', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        mocker = mock.Mock()
        mocker.return_value = 'abc123'
        with mock.patch("osc_lib.utils.get_password", mocker):
            columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'enabled': True,
            'tenant_id': None,
        }
        # UserManager.create(name, password, email, tenant_id=, enabled=)
        self.users_mock.create.assert_called_with(
            self.fake_user_c.name, 'abc123', None, **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_email(self):
        arglist = [
            '--email',
            'barney@example.com',
            self.fake_user_c.name,
        ]
        verifylist = [
            ('name', self.fake_user_c.name),
            ('email', 'barney@example.com'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'enabled': True,
            'tenant_id': None,
        }
        # UserManager.create(name, password, email, tenant_id=, enabled=)
        self.users_mock.create.assert_called_with(
            self.fake_user_c.name, None, 'barney@example.com', **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_project(self):
        # Return the new project
        self.projects_mock.get.return_value = self.fake_project_c

        # Set up to return an updated user
        attr = {
            'tenantId': self.fake_project_c.id,
        }
        user_2 = identity_fakes.FakeUser.create_one_user(attr)
        self.users_mock.create.return_value = user_2

        arglist = [
            '--project',
            self.fake_project_c.name,
            user_2.name,
        ]
        verifylist = [
            ('name', user_2.name),
            ('project', self.fake_project_c.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'enabled': True,
            'tenant_id': self.fake_project_c.id,
        }
        # UserManager.create(name, password, email, tenant_id=, enabled=)
        self.users_mock.create.assert_called_with(
            user_2.name, None, None, **kwargs
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            user_2.email,
            True,
            user_2.id,
            user_2.name,
            self.fake_project_c.id,
        )
        self.assertEqual(datalist, data)

    def test_user_create_enable(self):
        arglist = [
            '--enable',
            self.fake_user_c.name,
        ]
        verifylist = [
            ('name', self.fake_user_c.name),
            ('enable', True),
            ('disable', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'enabled': True,
            'tenant_id': None,
        }
        # UserManager.create(name, password, email, tenant_id=, enabled=)
        self.users_mock.create.assert_called_with(
            self.fake_user_c.name, None, None, **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_disable(self):
        arglist = [
            '--disable',
            self.fake_user_c.name,
        ]
        verifylist = [
            ('name', self.fake_user_c.name),
            ('enable', False),
            ('disable', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'enabled': False,
            'tenant_id': None,
        }
        # UserManager.create(name, password, email, tenant_id=, enabled=)
        self.users_mock.create.assert_called_with(
            self.fake_user_c.name, None, None, **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_or_show_exists(self):
        def _raise_conflict(*args, **kwargs):
            raise ks_exc.Conflict(None)

        # need to make this throw an exception...
        self.users_mock.create.side_effect = _raise_conflict

        self.users_mock.get.return_value = self.fake_user_c

        arglist = [
            '--or-show',
            self.fake_user_c.name,
        ]
        verifylist = [
            ('name', self.fake_user_c.name),
            ('or_show', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # UserManager.create(name, password, email, tenant_id=, enabled=)
        self.users_mock.get.assert_called_with(self.fake_user_c.name)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_user_create_or_show_not_exists(self):
        arglist = [
            '--or-show',
            self.fake_user_c.name,
        ]
        verifylist = [
            ('name', self.fake_user_c.name),
            ('or_show', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'enabled': True,
            'tenant_id': None,
        }
        # UserManager.create(name, password, email, tenant_id=, enabled=)
        self.users_mock.create.assert_called_with(
            self.fake_user_c.name, None, None, **kwargs
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)


class TestUserDelete(TestUser):
    def setUp(self):
        super().setUp()

        # This is the return value for utils.find_resource()
        self.users_mock.get.return_value = self.fake_user
        self.users_mock.delete.return_value = None

        # Get the command object to test
        self.cmd = user.DeleteUser(self.app, None)

    def test_user_delete_no_options(self):
        arglist = [
            self.fake_user.id,
        ]
        verifylist = [
            ('users', [self.fake_user.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.users_mock.delete.assert_called_with(
            self.fake_user.id,
        )
        self.assertIsNone(result)

    @mock.patch.object(utils, 'find_resource')
    def test_delete_multi_users_with_exception(self, find_mock):
        find_mock.side_effect = [self.fake_user, exceptions.CommandError]
        arglist = [
            self.fake_user.id,
            'unexist_user',
        ]
        verifylist = [
            ('users', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 users failed to delete.', str(e))

        find_mock.assert_any_call(self.users_mock, self.fake_user.id)
        find_mock.assert_any_call(self.users_mock, 'unexist_user')

        self.assertEqual(2, find_mock.call_count)
        self.users_mock.delete.assert_called_once_with(self.fake_user.id)


class TestUserList(TestUser):
    fake_project_l = identity_fakes.FakeProject.create_one_project()
    attr = {
        'tenantId': fake_project_l.id,
    }
    fake_user_l = identity_fakes.FakeUser.create_one_user(attr)

    columns = (
        'ID',
        'Name',
    )
    datalist = (
        (
            fake_user_l.id,
            fake_user_l.name,
        ),
    )

    def setUp(self):
        super().setUp()

        self.projects_mock.get.return_value = self.fake_project_l
        self.projects_mock.list.return_value = [self.fake_project_l]

        self.users_mock.list.return_value = [self.fake_user_l]

        # Get the command object to test
        self.cmd = user.ListUser(self.app, None)

    def test_user_list_no_options(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.users_mock.list.assert_called_with(tenant_id=None)

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.datalist, tuple(data))

    def test_user_list_project(self):
        arglist = [
            '--project',
            self.fake_project_l.id,
        ]
        verifylist = [
            ('project', self.fake_project_l.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        project_id = self.fake_project_l.id

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.users_mock.list.assert_called_with(tenant_id=project_id)

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.datalist, tuple(data))

    def test_user_list_long(self):
        arglist = [
            '--long',
        ]
        verifylist = [
            ('long', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.users_mock.list.assert_called_with(tenant_id=None)

        collist = ('ID', 'Name', 'Project', 'Email', 'Enabled')
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.fake_user_l.id,
                self.fake_user_l.name,
                user.ProjectColumn(
                    self.fake_project_l.id,
                    {self.fake_project_l.id: self.fake_project_l},
                ),
                self.fake_user_l.email,
                True,
            ),
        )
        self.assertCountEqual(datalist, tuple(data))


class TestUserSet(TestUser):
    def setUp(self):
        super().setUp()

        self.projects_mock.get.return_value = self.fake_project
        self.users_mock.get.return_value = self.fake_user

        # Get the command object to test
        self.cmd = user.SetUser(self.app, None)

    def test_user_set_no_options(self):
        arglist = [
            self.fake_user.name,
        ]
        verifylist = [
            ('name', None),
            ('password', None),
            ('email', None),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.fake_user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)

    def test_user_set_unexist_user(self):
        arglist = [
            "unexist-user",
        ]
        verifylist = [
            ('name', None),
            ('password', None),
            ('email', None),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', "unexist-user"),
        ]
        self.users_mock.get.side_effect = exceptions.NotFound(None)
        self.users_mock.find.side_effect = exceptions.NotFound(None)

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_user_set_name(self):
        arglist = [
            '--name',
            'qwerty',
            self.fake_user.name,
        ]
        verifylist = [
            ('name', 'qwerty'),
            ('password', None),
            ('email', None),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.fake_user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'enabled': True,
            'name': 'qwerty',
        }
        # UserManager.update(user, **kwargs)
        self.users_mock.update.assert_called_with(self.fake_user.id, **kwargs)
        self.assertIsNone(result)

    def test_user_set_password(self):
        arglist = [
            '--password',
            'secret',
            self.fake_user.name,
        ]
        verifylist = [
            ('name', None),
            ('password', 'secret'),
            ('password_prompt', False),
            ('email', None),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.fake_user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # UserManager.update_password(user, password)
        self.users_mock.update_password.assert_called_with(
            self.fake_user.id,
            'secret',
        )
        self.assertIsNone(result)

    def test_user_set_password_prompt(self):
        arglist = [
            '--password-prompt',
            self.fake_user.name,
        ]
        verifylist = [
            ('name', None),
            ('password', None),
            ('password_prompt', True),
            ('email', None),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.fake_user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        mocker = mock.Mock()
        mocker.return_value = 'abc123'
        with mock.patch("osc_lib.utils.get_password", mocker):
            result = self.cmd.take_action(parsed_args)

        # UserManager.update_password(user, password)
        self.users_mock.update_password.assert_called_with(
            self.fake_user.id,
            'abc123',
        )
        self.assertIsNone(result)

    def test_user_set_email(self):
        arglist = [
            '--email',
            'barney@example.com',
            self.fake_user.name,
        ]
        verifylist = [
            ('name', None),
            ('password', None),
            ('email', 'barney@example.com'),
            ('project', None),
            ('enable', False),
            ('disable', False),
            ('user', self.fake_user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'email': 'barney@example.com',
            'enabled': True,
        }
        # UserManager.update(user, **kwargs)
        self.users_mock.update.assert_called_with(self.fake_user.id, **kwargs)
        self.assertIsNone(result)

    def test_user_set_project(self):
        arglist = [
            '--project',
            self.fake_project.id,
            self.fake_user.name,
        ]
        verifylist = [
            ('name', None),
            ('password', None),
            ('email', None),
            ('project', self.fake_project.id),
            ('enable', False),
            ('disable', False),
            ('user', self.fake_user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # UserManager.update_tenant(user, tenant)
        self.users_mock.update_tenant.assert_called_with(
            self.fake_user.id,
            self.fake_project.id,
        )
        self.assertIsNone(result)

    def test_user_set_enable(self):
        arglist = [
            '--enable',
            self.fake_user.name,
        ]
        verifylist = [
            ('name', None),
            ('password', None),
            ('email', None),
            ('project', None),
            ('enable', True),
            ('disable', False),
            ('user', self.fake_user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'enabled': True,
        }
        # UserManager.update(user, **kwargs)
        self.users_mock.update.assert_called_with(self.fake_user.id, **kwargs)
        self.assertIsNone(result)

    def test_user_set_disable(self):
        arglist = [
            '--disable',
            self.fake_user.name,
        ]
        verifylist = [
            ('name', None),
            ('password', None),
            ('email', None),
            ('project', None),
            ('enable', False),
            ('disable', True),
            ('user', self.fake_user.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'enabled': False,
        }
        # UserManager.update(user, **kwargs)
        self.users_mock.update.assert_called_with(self.fake_user.id, **kwargs)
        self.assertIsNone(result)


class TestUserShow(TestUser):
    def setUp(self):
        super().setUp()

        self.users_mock.get.return_value = self.fake_user

        # Get the command object to test
        self.cmd = user.ShowUser(self.app, None)

    def test_user_show(self):
        arglist = [
            self.fake_user.id,
        ]
        verifylist = [
            ('user', self.fake_user.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        self.users_mock.get.assert_called_with(self.fake_user.id)

        collist = ('email', 'enabled', 'id', 'name', 'project_id')
        self.assertEqual(collist, columns)
        datalist = (
            self.fake_user.email,
            True,
            self.fake_user.id,
            self.fake_user.name,
            self.fake_project.id,
        )
        self.assertCountEqual(datalist, data)



# Copyright 2017 FUJITSU LIMITED
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from openstack.identity.v3 import project as _project
from openstack.network.v2 import vpn_ipsec_policy as sdk_vpn_ipsecp_p
from openstack.test import fakes as sdk_fakes
from openstackclient.tests.unit import utils as tests_utils
from osc_lib import exceptions

from openstackclient.network.v2.vpnaas import ipsecpolicy
from openstackclient.tests.unit.network.v2 import fakes as test_fakes
from openstackclient.tests.unit.network.v2.vpnaas import fakes


class TestIPSecPolicy(test_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()

        self.project = sdk_fakes.generate_fake_resource(_project.Project)
        self.identity_sdk_client.find_project.return_value = self.project

        self._ipsecpolicy = fakes.IPSecPolicy().create()

        def _mock_ipsecpolicy(*args, **kwargs):
            self.network_client.find_vpn_ipsec_policy.assert_called_once_with(
                self._ipsecpolicy['id'], ignore_missing=False
            )
            return {'id': args[0]}

        self.network_client.find_vpn_ipsec_policy.side_effect = mock.Mock(
            side_effect=_mock_ipsecpolicy
        )
        self.headers = (
            'ID',
            'Name',
            'Authentication Algorithm',
            'Encapsulation Mode',
            'Transform Protocol',
            'Encryption Algorithm',
            'Perfect Forward Secrecy (PFS)',
            'Description',
            'Project',
            'Lifetime',
        )
        self.ordered_headers = (
            'Authentication Algorithm',
            'Description',
            'Encapsulation Mode',
            'Encryption Algorithm',
            'ID',
            'Lifetime',
            'Name',
            'Perfect Forward Secrecy (PFS)',
            'Project',
            'Transform Protocol',
        )
        self.ordered_data = (
            self._ipsecpolicy['auth_algorithm'],
            self._ipsecpolicy['description'],
            self._ipsecpolicy['encapsulation_mode'],
            self._ipsecpolicy['encryption_algorithm'],
            self._ipsecpolicy['id'],
            self._ipsecpolicy['lifetime'],
            self._ipsecpolicy['name'],
            self._ipsecpolicy['pfs'],
            self._ipsecpolicy['project_id'],
            self._ipsecpolicy['transform_protocol'],
        )
        self.ordered_columns = (
            'auth_algorithm',
            'description',
            'encapsulation_mode',
            'encryption_algorithm',
            'id',
            'lifetime',
            'name',
            'pfs',
            'project_id',
            'transform_protocol',
        )


class TestCreateIPSecPolicy(TestIPSecPolicy):
    def setUp(self):
        super().setUp()

        self.project = sdk_fakes.generate_fake_resource(_project.Project)
        self.identity_sdk_client.find_project.return_value = self.project

        self.network_client.create_vpn_ipsec_policy.return_value = (
            self._ipsecpolicy
        )
        self.mocked = self.network_client.create_vpn_ipsec_policy
        self.cmd = ipsecpolicy.CreateIPsecPolicy(self.app, None)

    def test_create_with_no_options(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_create_with_all_params(self):
        arglist = [
            'my-name',
            '--auth-algorithm',
            'sha1',
            '--encapsulation-mode',
            'tunnel',
            '--transform-protocol',
            'esp',
            '--encryption-algorithm',
            'aes-128',
            '--pfs',
            'group5',
            '--description',
            'my-desc',
            '--project',
            self.project.id,
        ]
        verifylist = [
            ('name', 'my-name'),
            ('auth_algorithm', 'sha1'),
            ('encapsulation_mode', 'tunnel'),
            ('transform_protocol', 'esp'),
            ('encryption_algorithm', 'aes-128'),
            ('pfs', 'group5'),
            ('description', 'my-desc'),
            ('project', self.project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)
        self.network_client.create_vpn_ipsec_policy.assert_called_once_with(
            name='my-name',
            auth_algorithm='sha1',
            encapsulation_mode='tunnel',
            transform_protocol='esp',
            encryption_algorithm='aes-128',
            pfs='group5',
            description='my-desc',
            project_id=self.project.id,
        )
        self.assertEqual(self.ordered_headers, tuple(sorted(headers)))
        self.assertCountEqual(self.ordered_data, data)

    def test_create_with_all_params_name(self):
        arglist = [
            'new_ipsecpolicy',
            '--auth-algorithm',
            'sha1',
            '--encapsulation-mode',
            'tunnel',
            '--transform-protocol',
            'esp',
            '--encryption-algorithm',
            'aes-128',
            '--pfs',
            'group5',
            '--description',
            'my-desc',
            '--project',
            self.project.id,
        ]
        verifylist = [
            ('name', 'new_ipsecpolicy'),
            ('auth_algorithm', 'sha1'),
            ('encapsulation_mode', 'tunnel'),
            ('transform_protocol', 'esp'),
            ('encryption_algorithm', 'aes-128'),
            ('pfs', 'group5'),
            ('description', 'my-desc'),
            ('project', self.project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)
        self.network_client.create_vpn_ipsec_policy.assert_called_once_with(
            name='new_ipsecpolicy',
            auth_algorithm='sha1',
            encapsulation_mode='tunnel',
            transform_protocol='esp',
            encryption_algorithm='aes-128',
            pfs='group5',
            description='my-desc',
            project_id=self.project.id,
        )
        self.assertEqual(self.ordered_headers, tuple(sorted(headers)))
        self.assertCountEqual(self.ordered_data, data)


class TestDeleteIPSecPolicy(TestIPSecPolicy):
    def setUp(self):
        super().setUp()
        self.cmd = ipsecpolicy.DeleteIPsecPolicy(self.app, None)

    def test_delete_with_one_resource(self):
        def _mock_ipsec_p(*args, **kwargs):
            return sdk_vpn_ipsecp_p.VpnIpsecPolicy(**{'id': args[0]})

        self.network_client.find_vpn_ipsec_policy.side_effect = _mock_ipsec_p

        arglist = [self._ipsecpolicy['id']]
        verifylist = [('ipsecpolicy', [self._ipsecpolicy['id']])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.delete_vpn_ipsec_policy.assert_called_once_with(
            self._ipsecpolicy['id']
        )
        self.assertIsNone(result)

    def test_delete_with_multiple_resources(self):

        def _mock_ipsec_p(*args, **kwargs):
            return sdk_vpn_ipsecp_p.VpnIpsecPolicy(**{'id': args[0]})

        self.network_client.find_vpn_ipsec_policy.side_effect = _mock_ipsec_p

        arglist = ['target1', 'target2']
        verifylist = [('ipsecpolicy', ['target1', 'target2'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.assertEqual(
            2, self.network_client.delete_vpn_ipsec_policy.call_count
        )
        for idx, reference in enumerate(['target1', 'target2']):
            actual = ''.join(
                self.network_client.delete_vpn_ipsec_policy.call_args_list[
                    idx
                ][0]
            )
            self.assertEqual(reference, actual)

    def test_delete_multiple_with_exception(self):
        arglist = ['target']
        verifylist = [('ipsecpolicy', ['target'])]

        self.network_client.find_vpn_ipsec_policy.side_effect = [
            'target',
            exceptions.CommandError,
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )


class TestListIPSecPolicy(TestIPSecPolicy):
    def setUp(self):
        super().setUp()
        self.cmd = ipsecpolicy.ListIPsecPolicy(self.app, None)

        self.short_header = (
            'ID',
            'Name',
            'Authentication Algorithm',
            'Encapsulation Mode',
            'Transform Protocol',
            'Encryption Algorithm',
        )

        self.short_data = (
            self._ipsecpolicy['id'],
            self._ipsecpolicy['name'],
            self._ipsecpolicy['auth_algorithm'],
            self._ipsecpolicy['encapsulation_mode'],
            self._ipsecpolicy['transform_protocol'],
            self._ipsecpolicy['encryption_algorithm'],
        )

        self.network_client.vpn_ipsec_policies.return_value = [
            self._ipsecpolicy
        ]

    def test_list_with_long_option(self):
        arglist = ['--long']
        verifylist = [('long', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, _data = self.cmd.take_action(parsed_args)

        self.network_client.vpn_ipsec_policies.assert_called_once_with()
        self.assertEqual(list(self.headers), headers)

    def test_list_with_no_option(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.network_client.vpn_ipsec_policies.assert_called_once_with()
        self.assertEqual(list(self.short_header), headers)
        self.assertEqual([self.short_data], list(data))


class TestSetIPSecPolicy(TestIPSecPolicy):
    def setUp(self):
        super().setUp()
        self.network_client.update_vpn_ipsec_policy.return_value = (
            self._ipsecpolicy
        )
        self.cmd = ipsecpolicy.SetIPsecPolicy(self.app, None)

    def test_set_auth_algorithm_with_sha256(self):
        arglist = [self._ipsecpolicy['id'], '--auth-algorithm', 'sha256']
        verifylist = [
            ('ipsecpolicy', self._ipsecpolicy['id']),
            ('auth_algorithm', 'sha256'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_vpn_ipsec_policy.assert_called_once_with(
            self._ipsecpolicy['id'], auth_algorithm='sha256'
        )
        self.assertIsNone(result)

    def test_set_name(self):
        arglist = [self._ipsecpolicy['id'], '--name', 'change']
        verifylist = [
            ('ipsecpolicy', self._ipsecpolicy['id']),
            ('name', 'change'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_vpn_ipsec_policy.assert_called_once_with(
            self._ipsecpolicy['id'], name='change'
        )
        self.assertIsNone(result)

    def test_set_description(self):
        arglist = [self._ipsecpolicy['id'], '--description', 'change-desc']
        verifylist = [
            ('ipsecpolicy', self._ipsecpolicy['id']),
            ('description', 'change-desc'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_vpn_ipsec_policy.assert_called_once_with(
            self._ipsecpolicy['id'], description='change-desc'
        )
        self.assertIsNone(result)


class TestShowIPSecPolicy(TestIPSecPolicy):
    def setUp(self):
        super().setUp()

        self.network_client.find_vpn_ipsec_policy.side_effect = None
        self.network_client.find_vpn_ipsec_policy.return_value = (
            self._ipsecpolicy
        )
        self.cmd = ipsecpolicy.ShowIPsecPolicy(self.app, None)

    def test_show_filtered_by_id_or_name(self):
        arglist = [self._ipsecpolicy['id']]
        verifylist = [('ipsecpolicy', self._ipsecpolicy['id'])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.network_client.find_vpn_ipsec_policy.assert_called_once_with(
            self._ipsecpolicy['id'], ignore_missing=False
        )
        self.assertEqual(self.ordered_headers, headers)
        self.assertCountEqual(self.ordered_data, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock

from openstackclient.identity.v3 import endpoint_group
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes


class TestEndpointGroup(identity_fakes.TestIdentity):
    def setUp(self):
        super().setUp()

        # Get a shortcut to the EndpointManager Mock
        self.endpoint_groups_mock = self.identity_client.endpoint_groups
        self.endpoint_groups_mock.reset_mock()
        self.epf_mock = self.identity_client.endpoint_filter
        self.epf_mock.reset_mock()

        # Get a shortcut to the ServiceManager Mock
        self.services_mock = self.identity_client.services
        self.services_mock.reset_mock()

        # Get a shortcut to the DomainManager Mock
        self.domains_mock = self.identity_client.domains
        self.domains_mock.reset_mock()

        # Get a shortcut to the ProjectManager Mock
        self.projects_mock = self.identity_client.projects
        self.projects_mock.reset_mock()


class TestEndpointGroupCreate(TestEndpointGroup):
    columns = (
        'description',
        'filters',
        'id',
        'name',
    )

    def setUp(self):
        super().setUp()

        self.endpoint_group = (
            identity_fakes.FakeEndpointGroup.create_one_endpointgroup(
                attrs={'filters': identity_fakes.endpoint_group_filters}
            )
        )

        self.endpoint_groups_mock.create.return_value = self.endpoint_group

        # Get the command object to test
        self.cmd = endpoint_group.CreateEndpointGroup(self.app, None)

    def test_endpointgroup_create_no_options(self):
        arglist = [
            '--description',
            self.endpoint_group.description,
            self.endpoint_group.name,
            identity_fakes.endpoint_group_file_path,
        ]
        verifylist = [
            ('name', self.endpoint_group.name),
            ('filters', identity_fakes.endpoint_group_file_path),
            ('description', self.endpoint_group.description),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        mocker = mock.Mock()
        mocker.return_value = identity_fakes.endpoint_group_filters
        with mock.patch(
            "openstackclient.identity.v3.endpoint_group."
            "CreateEndpointGroup._read_filters",
            mocker,
        ):
            columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.endpoint_group.name,
            'filters': identity_fakes.endpoint_group_filters,
            'description': self.endpoint_group.description,
        }

        self.endpoint_groups_mock.create.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            self.endpoint_group.description,
            identity_fakes.endpoint_group_filters,
            self.endpoint_group.id,
            self.endpoint_group.name,
        )
        self.assertEqual(datalist, data)


class TestEndpointGroupDelete(TestEndpointGroup):
    endpoint_group = (
        identity_fakes.FakeEndpointGroup.create_one_endpointgroup()
    )

    def setUp(self):
        super().setUp()

        # This is the return value for utils.find_resource(endpoint)
        self.endpoint_groups_mock.get.return_value = self.endpoint_group
        self.endpoint_groups_mock.delete.return_value = None

        # Get the command object to test
        self.cmd = endpoint_group.DeleteEndpointGroup(self.app, None)

    def test_endpointgroup_delete(self):
        arglist = [
            self.endpoint_group.id,
        ]
        verifylist = [
            ('endpointgroup', [self.endpoint_group.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.endpoint_groups_mock.delete.assert_called_with(
            self.endpoint_group.id,
        )
        self.assertIsNone(result)


class TestEndpointGroupList(TestEndpointGroup):
    endpoint_group = (
        identity_fakes.FakeEndpointGroup.create_one_endpointgroup()
    )
    project = identity_fakes.FakeProject.create_one_project()
    domain = identity_fakes.FakeDomain.create_one_domain()

    columns = (
        'ID',
        'Name',
        'Description',
    )

    def setUp(self):
        super().setUp()

        self.endpoint_groups_mock.list.return_value = [self.endpoint_group]
        self.endpoint_groups_mock.get.return_value = self.endpoint_group
        self.epf_mock.list_projects_for_endpoint_group.return_value = [
            self.project
        ]
        self.epf_mock.list_endpoint_groups_for_project.return_value = [
            self.endpoint_group
        ]

        # Get the command object to test
        self.cmd = endpoint_group.ListEndpointGroup(self.app, None)

    def test_endpoint_group_list_no_options(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)
        self.endpoint_groups_mock.list.assert_called_with()

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                self.endpoint_group.id,
                self.endpoint_group.name,
                self.endpoint_group.description,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_endpoint_group_list_projects_by_endpoint_group(self):
        arglist = [
            '--endpointgroup',
            self.endpoint_group.id,
        ]
        verifylist = [
            ('endpointgroup', self.endpoint_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)
        self.epf_mock.list_projects_for_endpoint_group.assert_called_with(
            endpoint_group=self.endpoint_group.id
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                self.project.id,
                self.project.name,
                self.project.description,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_endpoint_group_list_by_project(self):
        self.epf_mock.list_endpoints_for_project.return_value = [
            self.endpoint_group
        ]
        self.projects_mock.get.return_value = self.project

        arglist = [
            '--project',
            self.project.name,
            '--domain',
            self.domain.name,
        ]
        verifylist = [
            ('project', self.project.name),
            ('domain', self.domain.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)
        self.epf_mock.list_endpoint_groups_for_project.assert_called_with(
            project=self.project.id
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            (
                self.endpoint_group.id,
                self.endpoint_group.name,
                self.endpoint_group.description,
            ),
        )
        self.assertEqual(datalist, tuple(data))


class TestEndpointGroupSet(TestEndpointGroup):
    endpoint_group = (
        identity_fakes.FakeEndpointGroup.create_one_endpointgroup()
    )

    def setUp(self):
        super().setUp()

        # This is the return value for utils.find_resource(endpoint)
        self.endpoint_groups_mock.get.return_value = self.endpoint_group

        self.endpoint_groups_mock.update.return_value = self.endpoint_group

        # Get the command object to test
        self.cmd = endpoint_group.SetEndpointGroup(self.app, None)

    def test_endpoint_group_set_no_options(self):
        arglist = [
            self.endpoint_group.id,
        ]
        verifylist = [
            ('endpointgroup', self.endpoint_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        kwargs = {'name': None, 'filters': None, 'description': ''}
        self.endpoint_groups_mock.update.assert_called_with(
            self.endpoint_group.id, **kwargs
        )
        self.assertIsNone(result)

    def test_endpoint_group_set_name(self):
        arglist = ['--name', 'qwerty', self.endpoint_group.id]
        verifylist = [
            ('name', 'qwerty'),
            ('endpointgroup', self.endpoint_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {'name': 'qwerty', 'filters': None, 'description': ''}
        self.endpoint_groups_mock.update.assert_called_with(
            self.endpoint_group.id, **kwargs
        )
        self.assertIsNone(result)

    def test_endpoint_group_set_filters(self):
        arglist = [
            '--filters',
            identity_fakes.endpoint_group_file_path,
            self.endpoint_group.id,
        ]
        verifylist = [
            ('filters', identity_fakes.endpoint_group_file_path),
            ('endpointgroup', self.endpoint_group.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        mocker = mock.Mock()
        mocker.return_value = identity_fakes.endpoint_group_filters_2
        with mock.patch(
            "openstackclient.identity.v3.endpoint_group."
            "SetEndpointGroup._read_filters",
            mocker,
        ):
            result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': None,
            'filters': identity_fakes.endpoint_group_filters_2,
            'description': '',
        }

        self.endpoint_groups_mock.update.assert_called_with(
            self.endpoint_group.id, **kwargs
        )

        self.assertIsNone(result)

    def test_endpoint_group_set_description(self):
        arglist = ['--description', 'qwerty', self.endpoint_group.id]
        verifylist = [
            ('description', 'qwerty'),
            ('endpointgroup', self.endpoint_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': None,
            'filters': None,
            'description': 'qwerty',
        }
        self.endpoint_groups_mock.update.assert_called_with(
            self.endpoint_group.id, **kwargs
        )
        self.assertIsNone(result)


class TestAddProjectToEndpointGroup(TestEndpointGroup):
    project = identity_fakes.FakeProject.create_one_project()
    domain = identity_fakes.FakeDomain.create_one_domain()
    endpoint_group = (
        identity_fakes.FakeEndpointGroup.create_one_endpointgroup()
    )

    new_ep_filter = (
        identity_fakes.FakeEndpointGroup.create_one_endpointgroup_filter(
            attrs={'endpointgroup': endpoint_group.id, 'project': project.id}
        )
    )

    def setUp(self):
        super().setUp()

        # This is the return value for utils.find_resource()
        self.endpoint_groups_mock.get.return_value = self.endpoint_group

        # Update the image_id in the MEMBER dict
        self.epf_mock.create.return_value = self.new_ep_filter
        self.projects_mock.get.return_value = self.project
        self.domains_mock.get.return_value = self.domain

        # Get the command object to test
        self.cmd = endpoint_group.AddProjectToEndpointGroup(self.app, None)

    def test_add_project_to_endpoint_no_option(self):
        arglist = [
            self.endpoint_group.id,
            self.project.id,
        ]
        verifylist = [
            ('endpointgroup', self.endpoint_group.id),
            ('project', self.project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.epf_mock.add_endpoint_group_to_project.assert_called_with(
            project=self.project.id,
            endpoint_group=self.endpoint_group.id,
        )
        self.assertIsNone(result)

    def test_add_project_to_endpoint_with_option(self):
        arglist = [
            self.endpoint_group.id,
            self.project.id,
            '--project-domain',
            self.domain.id,
        ]
        verifylist = [
            ('endpointgroup', self.endpoint_group.id),
            ('project', self.project.id),
            ('project_domain', self.domain.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.epf_mock.add_endpoint_group_to_project.assert_called_with(
            project=self.project.id,
            endpoint_group=self.endpoint_group.id,
        )
        self.assertIsNone(result)


class TestRemoveProjectEndpointGroup(TestEndpointGroup):
    project = identity_fakes.FakeProject.create_one_project()
    domain = identity_fakes.FakeDomain.create_one_domain()
    endpoint_group = (
        identity_fakes.FakeEndpointGroup.create_one_endpointgroup()
    )

    def setUp(self):
        super().setUp()

        # This is the return value for utils.find_resource()
        self.endpoint_groups_mock.get.return_value = self.endpoint_group

        self.projects_mock.get.return_value = self.project
        self.domains_mock.get.return_value = self.domain
        self.epf_mock.delete.return_value = None

        # Get the command object to test
        self.cmd = endpoint_group.RemoveProjectFromEndpointGroup(
            self.app, None
        )

    def test_remove_project_endpoint_no_options(self):
        arglist = [
            self.endpoint_group.id,
            self.project.id,
        ]
        verifylist = [
            ('endpointgroup', self.endpoint_group.id),
            ('project', self.project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.epf_mock.delete_endpoint_group_from_project.assert_called_with(
            project=self.project.id,
            endpoint_group=self.endpoint_group.id,
        )
        self.assertIsNone(result)

    def test_remove_project_endpoint_with_options(self):
        arglist = [
            self.endpoint_group.id,
            self.project.id,
            '--project-domain',
            self.domain.id,
        ]
        verifylist = [
            ('endpointgroup', self.endpoint_group.id),
            ('project', self.project.id),
            ('project_domain', self.domain.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.epf_mock.delete_endpoint_group_from_project.assert_called_with(
            project=self.project.id,
            endpoint_group=self.endpoint_group.id,
        )
        self.assertIsNone(result)
